@@ -167,7 +167,7 @@ def bic_g(data, cols, weights=None):
 # compute weights of data with missing values
 def compute_weights(data, varnames, W_ids, cause_list):
     if 'int' in data.dtype.name:
-        data_delete = data[data[:, W_ids].min(axis=1) >= 0]
+        data_delete = data[(data[:, W_ids].min(axis=1) >= 0) & (data[:, W_ids].max(axis=1) != np.iinfo(data.dtype).max)]
     else:
         data_delete = data[~np.isnan(data[:, W_ids]).any(axis=1)]
     weights = np.ones(len(data_delete))
